(*****************************************************************************
 * HOL-TestGen --- theorem-prover based test case generation
 *                 http://www.brucker.ch/projects/hol-testgen/
 *
 * MyKeOS.thy --- a Study for OS opertions, modelling atomic actions
 *                for sequence testing based on a 
 *                deterministic specification.
 *
 * This file is part of HOL-TestGen.
 *
 * Copyright (c) 2016 Univ. Paris-Sud, France
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *
 *     * Neither the name of the copyright holders nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************)

header{* The MyKeOS Case Study *}

theory  MyKeOS
imports 
   (* Testing *) 
   "../../../src/Testing" 
   (*"../../../src/codegen_fsharp/Code_Integer_Fsharp"*) 
begin
(*
declare [[testgen_profiling]]
*)

text{* This example is drawn from the operating system testing domain; it is a rough abstraction
of PiKeOS and explains the underlying techniques of this particular case study on a small example.
The full paper can be found under \cite{DBLP:conf/vstte/BruckerHNW15}. *}

 
text{* This is a fun-operating system --- closely following the Bank example ---
intended to explain the principles of symbolic execution used in our PikeOS study.

Moreover, in this scenario, we assume that the system under test is
deterministic.
 *}

text{* The state of a thread (belonging to a task, \ie{} a Unix/PosiX like ``process'' 
just modeled by a map from task-id/thread-id information to the number of a resource
(a communication channel descriptor, for example) that was allocated to a thread. *}
type_synonym task_id = int

type_synonym thread_id = int

type_synonym thread_local_var_tab = "(task_id \<times> thread_id) \<rightharpoonup> int"


subsubsection{* Operation definitions *}

text {* A standard, JML or OCL or VCC like interface specification
might look like:

\begin{verbatim}

Init:  forall (c,no) : dom(data_base::thread_local_var_tab). data_base(c,no)>=0

op alloc (c : task_id, no : thread_id, amount:nat) : unit
pre  (c,no) : dom(data_base)
post data_base'=data_base[(c,no) := data_base(c,no) + amount]

op release(c : task_id, no : thread_id, amount:nat) : unit
pre  (c,no) : dom(data_base) and data_base(c,no) >= amount
post data_base'=data_base[(c,no) := data_base(c,no) - amount]

op status (c : task_id, no : thread_id) : int
pre  (c,no) : dom(data_base)
post data_base'=data_base and result = data_base(c,no)


\end{verbatim}

*}



text{* Interface normalization turns this interface into the following input type: *}

datatype in_c = alloc   task_id thread_id nat
              | release task_id thread_id nat
              | status  task_id thread_id
              
typ "MyKeOS.in_c"

datatype out_c = alloc_ok | release_ok | status_ok nat 

fun    precond :: "thread_local_var_tab \<Rightarrow> in_c \<Rightarrow> bool"
where "precond \<sigma> (alloc c no m) = ((c,no) \<in> dom \<sigma>)"
    | "precond \<sigma> (release c no m) = ((c,no) \<in> dom \<sigma> \<and> (int m) \<le> the(\<sigma>(c,no)))"
    | "precond \<sigma> (status c no) = ((c,no) \<in> dom \<sigma>)"

fun    postcond :: "in_c \<Rightarrow> thread_local_var_tab \<Rightarrow> (out_c \<times> thread_local_var_tab) set"
where "postcond (alloc c no m) \<sigma> =
         { (n,\<sigma>'). (n = alloc_ok \<and> \<sigma>'=\<sigma>((c,no)\<mapsto> the(\<sigma>(c,no)) + int m))}"
    | "postcond (release c no m) \<sigma> =
         { (n,\<sigma>'). (n = release_ok \<and> \<sigma>'=\<sigma>((c,no)\<mapsto> the(\<sigma>(c,no)) - int m))}"
    | "postcond (status c no) \<sigma> =
         { (n,\<sigma>'). (\<sigma>=\<sigma>' \<and> (\<exists> x. status_ok x = n \<and> x = nat(the(\<sigma>(c,no)))))}"




subsubsection{* Constructing an Abstract Program *}

text{* Using the Operators \verb+impl+ and \verb+strong_impl+, we can 
synthesize an abstract program right away from the specification, i.e. the pair 
of pre- and  postcondition defined above. Since this program is even 
deterministic, we derive a set of symbolic execution rules used in the test case 
generation process which will produce symbolic results against which the PUT can 
be compared in the test driver. *}
(*
definition implementable :: "['\<sigma> \<Rightarrow> '\<iota> \<Rightarrow> bool,'\<iota> \<Rightarrow> ('o,'\<sigma>)MON_SB] \<Rightarrow> bool"
where     "implementable pre post =(\<forall> \<sigma> \<iota>. pre \<sigma> \<iota> \<longrightarrow>(\<exists> out \<sigma>'. (out,\<sigma>') \<in> post \<iota> \<sigma> ))"
*)

lemma precond_postcond_implementable: 
     "implementable precond postcond"
apply(auto simp: implementable_def)
apply(case_tac "\<iota>", simp_all)
done

text{* Based on this machinery, it is now possible to construct the system model as the 
       canonical completion of the (functional) specification consisting of pre-  and 
       post-conditions*}

definition "SYS = (strong_impl precond postcond)"

lemma SYS_is_strong_impl : "is_strong_impl precond postcond SYS"
by(simp add: SYS_def is_strong_impl)

thm SYS_is_strong_impl[simplified is_strong_impl_def,THEN spec,of "(alloc c no m)", simplified, standard]
(*automatic, but logically not strong enough *)

subsubsection{* Proving Symbolic Execution Rules for the  Abstractly Program *}
text{* The following lemmas reveal that this "constructed" program is actually
(due to determinism of the spec): *}

lemma Eps_split_eq' : "(SOME (x', y'). x'= x  \<and> y'= y) = (SOME (x', y'). x = x' \<and> y = y')"
by(rule arg_cong[of _ _ "Eps"], auto)

interpretation alloc : efsm_det 
                      "precond" "postcond" "SYS" "(alloc c no m)" "\<lambda>_. alloc_ok" 
                      "\<lambda> \<sigma>. \<sigma>((c, no) \<mapsto>  (the(\<sigma>(c, no)) + int m))" "\<lambda> \<sigma>. ((c, no) \<in> dom \<sigma>)"
               by unfold_locales (auto simp: SYS_def Eps_split_eq')



interpretation release : efsm_det 
                        "precond" "postcond" "SYS" "(release c no m)" "\<lambda>_. release_ok" 
                         "\<lambda> \<sigma>. \<sigma>((c, no) \<mapsto>  (the(\<sigma>(c, no))-int m))" 
                         "\<lambda> \<sigma>.((c, no)\<in>dom \<sigma>) \<and> (int m)\<le>the(\<sigma>(c,no))"
               by unfold_locales (auto simp: SYS_def Eps_split_eq')

interpretation status : efsm_det 
                       "precond" "postcond" "SYS" "(status c no)" 
                       "\<lambda>\<sigma>. (status_ok (nat(the(\<sigma>(c, no)))))" 
                       "\<lambda> \<sigma>. \<sigma>" "\<lambda> \<sigma>. ((c, no) \<in> dom \<sigma>)"
               by unfold_locales (auto simp: SYS_def Eps_split_eq')




subsection{* Setup *}

text{* Now we close the theory of symbolic execution by \emph{exluding} elementary rewrite
steps on @{term mbind}, \ie{} the rules @{thm mbind.simps(1)} @{thm mbind.simps(2)} *}

declare mbind.simps(1) [simp del]
        mbind.simps(2) [simp del]

text{* Here comes an interesting detail revealing the power of the
approach: The generated sequences still respect the preconditions
imposed by the specification  - in this case, where we are talking about
a \verb$task_id$ for which a defined account exists and for which we will never
produce traces in which we release more money than available on it. *}

subsubsection{* Restricting the Test-Space by Test Purposes *}
text{* We introduce a constraint on the input sequence, in order to limit the test-space
       a little and eliminate logically possible, but irrelevant test-sequences for a specific
       test-purpose. In this case, we narrow down on test-sequences concerning a specific
       \verb$task_id$ @{term c} with a specific bank-account number @{term no}.
       
       We make the (in this case implicit, but as constraint explicitly stated) test hypothesis,
       that the @{term SUT} is correct if it behaves correct for a single \verb$task_id$.
       This boils down to the assumption that they are implemented as atomic transactions
       and interleaved processing does not interfere with a single thread. *}

fun test_purpose :: "[(task_id \<times> thread_id) list, in_c list] \<Rightarrow> bool"
where
  "test_purpose ((c,no)#R) [status c' no'] = ((c=c' \<and> no=no') \<or> test_purpose R [status c' no'])" 
| "test_purpose ((c,no)#R) ((alloc c' no' m)#S) = ((c=c' \<and> no=no' \<and> test_purpose ((c,no)#R) S)
                                                   \<or> test_purpose R ((alloc c' no' m)#S))"
| "test_purpose ((c,no)#R) ((release c' no' m)#S) = ((c=c' \<and> no=no' \<and> test_purpose ((c,no)#R) S)
                                                   \<or> test_purpose R ((release c' no' m)#S))"
| "test_purpose _ _ = False"

(* Definition buggy --- the R list must be separated into two parts, but one con not throw 
   away elements) *)

lemma [simp] : "test_purpose [] a = False" by simp
lemma [simp] : "test_purpose r [] = False" by simp

lemma [simp] : "test_purpose ((c,no)#R) [a] = ((a = status c no) \<or> test_purpose R [a])"
proof (induct "R")
  case Nil show ?case by(cases a, auto)
next
  case (Cons a' R') then show ?case 
       apply(cases a, simp_all)
       apply(cases "a'", simp)
       apply(cases "a'", simp)
       apply(cases "a'", simp)
       apply(rename_tac "int1" "int2" "aa" "b")
       apply(case_tac "c = int1 \<and> no = int2", auto)
       done
qed

find_theorems name:"in_c" name :"split"
lemma [simp] : "R\<noteq>[] \<Longrightarrow> test_purpose [(c,no),(c',no')] (a#R) = 
                          (((\<exists>m. a = (alloc c no m)) \<or> 
                            (\<exists>m. a = (release c no m)) \<or> 
                            (\<exists>m. a = (alloc c' no' m)) \<or> 
                            (\<exists>m. a = (release c' no' m))) 
                            \<and> test_purpose [(c,no),(c',no')] R) "
apply(simp add: List.neq_Nil_conv, elim exE,simp)
apply(auto split: in_c.split in_c.split_asm)
apply(cases a, auto) 
sorry

(* PUT must be a constant for coding conventions ... *)
consts PUT :: "in_c \<Rightarrow> '\<sigma> \<Rightarrow> (out_c \<times> '\<sigma>) option"


(* lemmas H1 = valid_mbind'_mtE  legacy *)

end
