(*****************************************************************************
 * HOL-TestGen --- theorem-prover based test case generation
 *                 http://www.brucker.ch/projects/hol-testgen/
 *
 * Permutations.thy --- a theory on (list) permutations.
 *
 * This file is part of HOL-TestGen.
 *
 * Copyright (c) 2009-15 Univ. Paris-Sud, France
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *
 *     * Neither the name of the copyright holders nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************)

theory  Permutations
imports Main
begin


section{* An Elementary theory on permutations *}



(* Simple lemmas about [List.member]. *)
lemma bang_member: "0\<le>i \<Longrightarrow> i<length S \<Longrightarrow> List.member S (S!i)"
proof(induct S arbitrary: i)
  case Nil then show ?case by simp
next
  case (Cons a S) then show ?case
  by (cases "i", simp_all add:member_rec)
qed

lemma member_index:
  "List.member S a \<Longrightarrow> \<exists>i. S!i = a"
proof(induct S)
  case Nil then show ?case
       by (simp add:List.member_def)
next
  case (Cons b S)
  have "a=b \<or> List.member S a" using Cons.prems by (simp add:List.member_def)
  moreover
  { assume "a=b"
    then have "(b#S)!0 = a" by simp
    then have ?case by (rule exI)
  }
  moreover
  { assume *: "List.member S a" (*then have ?case*)
(* Isabelle Club discussion
    apply(insert Cons.hyps * )
    apply simp
    apply(erule exE)
    apply(rename_tac i)
    thm exI
    apply(rule_tac x="Suc i" in exI)
    apply(simp)
*)
    then obtain i where "S!i = a" using Cons.hyps
      apply(simp)
      by(rule exE) (* Here, apply is not enough. What is the difference again? *)
    then have "(b#S)!(Suc i) = a" by simp
    then have ?case by (rule exI)
  }
  ultimately show ?case by blast
qed

lemma remove1_member: "List.member (remove1 b S) a \<Longrightarrow> List.member S a"
by(simp add:List.member_def List.set_remove1_subset Set.set_rev_mp)

lemma list_all_member: "list_all P l = (\<forall>a. List.member l a \<longrightarrow> P a)"
by (metis Ball_set_list_all in_set_member)


(* Simple lemmas about intervals of integers. *)
lemma range_append: "i\<le>k \<Longrightarrow> k\<le>j \<Longrightarrow> [i..<j] = [i..<k] @ [k..<j]"
by(induct "k-i" arbitrary:i, simp_all add:upt_conv_Cons)

lemma member_range: "(i \<le> k \<and> k < j) = List.member [i..<j] k"
by (simp add:List.member_def)

lemma member_range_remove1:
  "i\<le>k \<Longrightarrow> k<j \<Longrightarrow> remove1 k [i..<j] = [i..<k] @ [Suc k..<j]"
by(induct "k-i" arbitrary:i, simp_all add:upt_conv_Cons)


subsection{* A theory of permutations of lists *}


fun is_perm :: "'a list \<Rightarrow> 'a list \<Rightarrow> bool"
where "is_perm  []   [] = True"
    | "is_perm  []   T  = False"
    | is_perm_cons: "is_perm (a#S) T  = (if List.member T a
                                         then is_perm S (remove1 a T)
                                         else False)"

value "is_perm [1,2,3::int] [3,1,2]"
value "is_perm [1,2,3::int] [3,1,2,3]"

(* Elementary simplifications. *)
lemma is_perm_empty_T: "is_perm [] T \<Longrightarrow> T = []"
by(induct T, simp_all)

lemma is_perm_implies_same_length[simp]: "is_perm S T \<Longrightarrow> length S = length T" 
proof(induct S arbitrary: T) 
   case Nil then show ?case 
        by(simp add: is_perm_empty_T)
next
   case (Cons a S) then show ?case 
        apply(simp add:List.member_def List.length_remove1 split:split_if_asm)
        by (metis One_nat_def Suc_diff_1 length_pos_if_in_set)
qed


(* Relational reasoning on permutations: reflexivity, symmetry, transitivity,
  adding or removing an element, and elementary permutations. *)
lemma is_perm_refl[simp]: "is_perm S S" 
by(induct S, simp_all add:List.member_def)

(* In Coq standard library, this property is part of the definition of permutations. *)
lemma is_perm_skip: "is_perm (a#S) (a#T) = is_perm S T"
by(auto simp:List.member_def)


lemma is_perm_skip2: "is_perm (A @ S) (A @ T) = is_perm S T"
by(induct A, simp_all add:is_perm_skip del:is_perm_cons)


lemma is_perm_remove1_cons:
  "List.member T a \<Longrightarrow> is_perm (remove1 a T) S \<Longrightarrow> is_perm T (a # S)"
proof(induct T arbitrary: a S)
  case Nil assume "List.member [] a" then show ?case 
       by (simp add:List.member_def)
next
  case (Cons b T) then show ?case
       by(auto simp: List.member_def, simp split: split_if_asm)
qed

(* Isabelle Club discussion *)
lemma is_perm_remove1_bis:
  assumes "is_perm S T"
  shows "is_perm (remove1 a S) (remove1 a T)"
using assms
proof(induct S arbitrary: a T)
oops

lemma is_perm_remove1:
  "is_perm S T \<Longrightarrow> is_perm (remove1 a S) (remove1 a T)"
proof(induct S arbitrary: a T) print_cases
  case Nil have A: "T=[]" using Nil.prems is_perm_empty_T by (auto) 
           then show ?case 
       by( simp)
next
  case (Cons b S) then show ?case
       by(simp split:split_if_asm, metis in_set_remove1 List.member_def remove1_commute)
qed

lemma is_perm_upt_Suc_remove0:
  "is_perm [0 ..<Suc n] ip \<Longrightarrow> is_perm [Suc 0 ..<Suc n] (remove1 0 ip)"
by (metis is_perm_remove1 remove1.simps(2) upt_rec zero_less_Suc)

lemma is_perm_sym: "is_perm S T \<Longrightarrow> is_perm T S"
proof(induct S arbitrary: T) 
  case Nil 
       from Nil.prems 
       have "T = []" by (simp add:is_perm_empty_T)
       then show "is_perm T []" by simp
next
  case (Cons a S)
    (*   note XX = `is_perm (a # S) T` *)
       have *: "List.member T a" using Cons.prems by (simp split:split_if_asm)
       have "is_perm S (remove1 a T)" using Cons.prems by (simp split:split_if_asm)
       then have "is_perm (remove1 a T) S" using Cons.hyps by simp
       then show ?case using * by (simp add:is_perm_remove1_cons)
qed


lemma is_perm_member: "is_perm S T \<Longrightarrow> List.member S a \<Longrightarrow> List.member T a"
proof(induct S arbitrary: T)
  case Nil then show ?case by (simp add:List.member_def)
next
  case (Cons b S)
  assume IH: "\<And>T. is_perm S T \<Longrightarrow> List.member S a \<Longrightarrow> List.member T a"
 and H_perm: "is_perm (b # S) T"
   and H_in: "List.member (b # S) a"
  then show "List.member T a"
  proof -
    (* Proof by disjunction *)
    have "b=a \<or> List.member S a" using H_in by (simp add:member_rec)
    moreover
    (* Case 1 *)
    { assume "b=a" then have ?thesis using H_perm by (simp split:split_if_asm) }
    moreover
    (* Case 2 *)
    { assume "List.member S a" then have "List.member (remove1 b T) a" using H_perm and IH
        by (simp split:split_if_asm)
      then have ?thesis by (rule remove1_member)
    }
    (* Bringing them together *)
    ultimately have ?thesis .. (* by blast *)
    (* Then conclusion.
       Having to write that last line is a bit unsatisfactory. *)
    then show ?thesis .
  qed
qed

lemma is_perm_list_all: "is_perm S T \<Longrightarrow> list_all P S \<Longrightarrow> list_all P T"
by (metis list_all_member is_perm_member is_perm_sym)

lemma is_perm_trans: "is_perm S T \<Longrightarrow> is_perm T U \<Longrightarrow> is_perm S U"
proof(induct S arbitrary: T U)
  case Nil 
       have "T=[]" using Nil.prems is_perm_empty_T by auto
       then show ?case using Nil.prems by simp
next
  case (Cons a S) then show ?case
       by(simp split:split_if_asm,
          metis is_perm_member is_perm_remove1)
qed

lemma is_perm_swap: "is_perm (a#b#S) (b#a#S)"
by (simp add:member_rec)

lemma is_perm_permute_head_element: "is_perm (a#S@T) (S@a#T)"
proof(induct S)
  case Nil show ?case by simp
next
  case (Cons b S) from Cons.hyps show ?case
       by(metis append_Cons is_perm_trans is_perm_swap is_perm_skip)
qed


lemma is_perm_skip1: "is_perm (S@[a]) (T@[a]) = is_perm S T"
proof - 
  have *  : "is_perm (S@[a]) (T@[a]) = is_perm (a#S) (T@[a])" 
       by (metis append_Nil2 is_perm_permute_head_element is_perm_sym is_perm_trans)
  have ** : "... = is_perm (a#S) (a # T)" 
       by (metis append_Nil2 is_perm_permute_head_element is_perm_sym is_perm_trans)
  show ?thesis
       by (metis * ** is_perm_skip)
qed

lemma is_perm_skip3: "is_perm (S@A) (T@A) = is_perm S T"
proof (induct A arbitrary: S T) print_cases
  case Nil show ?case by simp
next
  case (Cons a A) have *: "S @ a # A = (S @ [a]) @ A" by (metis append_Cons append_Nil append_assoc)
                  have ** : "T @ a # A = (T @ [a]) @ A"by (metis append_Cons append_Nil append_assoc)

  show ?case
        by(subst *,subst **,simp only: Cons.hyps,simp add: is_perm_skip1)
qed


lemma is_perm_remove1_front:
  "is_perm S T \<Longrightarrow> List.member S a \<Longrightarrow> is_perm (a#(remove1 a S)) T"
by (metis is_perm_cons is_perm_member is_perm_remove1)


(* Interaction between [map] and [is_perm]. *)
lemma remove1_map_inj:
  "inj f \<Longrightarrow> remove1 (f a) (map f S) = map f (remove1 a S)"
proof(induct S)
  case Nil show ?case by simp
next
  case (Cons b S) then show ?case by (simp add:inj_on_def)
qed

lemma map_is_perm:
  "inj f \<Longrightarrow> is_perm (map f S) (map f T) \<Longrightarrow> is_perm S T"
proof(induct "S" arbitrary: T)
  case Nil
  then have "T=[]" by (metis Nil_is_map_conv is_perm_empty_T)
  then show ?case by (metis is_perm.simps(1))
next
  case (Cons a S)
  have "List.member T a" using Cons.prems
       apply(simp add:List.member_def split:split_if_asm)
       by(metis inj_image_mem_iff)
  then show ?case using Cons
       by(simp add:remove1_map_inj split:split_if_asm)
qed

lemma is_perm_map:
  "inj f \<Longrightarrow> is_perm S T \<Longrightarrow> is_perm (map f S) (map f T)"
proof(induct "S" arbitrary:T)
  case Nil
       have "T=[]" using Nil.prems by (simp add:is_perm_empty_T)
       then show ?case by simp
next
  case (Cons a S)
  have "List.member (map f T) (f a)" using Cons.prems
       by(simp add:List.member_def split:split_if_asm)
  then show ?case using Cons.prems Cons.hyps
       by(simp split:split_if_asm add:remove1_map_inj)
qed




(* The next goal is to get a function [ext_perm] that extends a permutations of
  indexes by inserting a new element [i] at the head. We first define a function
  [lift_from] that is used to update the tail of the list by shifting by one the
  elements that are greater than or equal to [i]. *)

definition lift_from :: "nat \<Rightarrow> nat \<Rightarrow> nat"
  where "lift_from i = (\<lambda>k. if k<i then k else Suc k)"

definition ext_perm :: "nat \<Rightarrow> nat list \<Rightarrow> nat list"
  where "ext_perm i p = i # map (lift_from i) p"


(* General lemmas about [lift_from]. *)
lemma lift_from_inj: "inj (lift_from i)"
by (simp add:inj_on_def lift_from_def)

lemma lift_from_mono: "j < k \<Longrightarrow> lift_from i j < lift_from i k"
by (metis less_SucI lift_from_def linorder_neqE_nat not_less_eq)


(* Lemmas about applying [lift_from] to an interval of integers. *)
lemma lift_from_upt_below: "j \<le> k \<Longrightarrow> map (lift_from k) [i..<j] = [i..<j]"
by(induct "j-i" arbitrary: i, simp_all add:upt_conv_Cons lift_from_def)

lemma lift_from_upt_above: "k \<le> i \<Longrightarrow> map (lift_from k) [i..<j] = [Suc i..<Suc j]"
      apply(induct "j-i" arbitrary: i, simp_all)
      apply(rule impI) 
      by(simp add:upt_conv_Cons lift_from_def)

lemma lift_from_upt: "i\<le>k \<Longrightarrow> k\<le>j \<Longrightarrow> map (lift_from k) [i..<j] = [i..<k] @ [Suc k..<Suc j]"
proof -
  assume H1: "i \<le> k" and H2: "k \<le> j"
  then show ?thesis
  proof -
    have "map (lift_from k) [i..<j] = map (lift_from k) ([i..<k] @ [k..<j])"
         using H1 H2 by (metis range_append)
    also have "... = [i..<k] @ [Suc k..<Suc j]"
         by(simp add:lift_from_upt_below lift_from_upt_above)
    finally show ?thesis .
  qed
qed


(* Proving that [ext_perm] applied to a permutation of indexes is still a permutation
  of indexes (provided the extension element is in the right range). *)

lemma ext_perm_correct:
  "0\<le>i \<Longrightarrow> i\<le>length p \<Longrightarrow> is_perm p [0..<length p] \<Longrightarrow> is_perm (ext_perm i p) [0..<Suc (length p)]"
proof -
  assume H1: "0\<le>i" and H2: "i\<le>length p" and H3: "is_perm p [0..<length p]"
  then show ?thesis
  proof -
    from H1 H2 have H4: "List.member [0..<Suc (length p)] i"
         by (metis le_imp_less_Suc member_range)
    then have "is_perm (map (lift_from i) p) (remove1 i [0..<Suc (length p)])" using H1 H2 H3
         apply(simp only:member_range_remove1 lift_from_upt[THEN HOL.sym])
         by(rule is_perm_map, rule lift_from_inj)
    then show ?thesis using H4
         by(simp only:ext_perm_def is_perm_cons split:split_if, auto)
  qed
qed


(* First occurrence of an element in a list. *)

primrec first_occ :: "'a \<Rightarrow> 'a list \<Rightarrow> nat"
where fo_cons: "first_occ a (b#S) = (if a=b then 0 else Suc (first_occ a S))"

(* A few lemmas on first occurrences. *)
lemma first_occ_member:
  "List.member S a \<Longrightarrow> S!(first_occ a S) = a"
proof(induct S)
  case Nil then show ?case by (simp add:List.member_def)
next
  case (Cons b S) then show ?case by (simp, metis member_rec(1))
qed

lemma first_occ_member_length:
  "List.member S a \<Longrightarrow> first_occ a S < length S"
proof(induct S)
  case Nil then show ?case by (simp add:List.member_def)
next
  case (Cons b S) then show ?case by (simp, metis member_rec(1))
qed

lemma remove1_first_occ:
  "List.member S a \<Longrightarrow> remove1 a S = take (first_occ a S) S @ drop (Suc (first_occ a S)) S"
proof(induct S)
  case Nil then show ?case by simp
next
  case (Cons b S) then show ?case by (simp, metis member_rec(1))
qed


(* A few lemmas on map. *)
lemma map_comp:
  "is_perm ip1 [0 ..<length S] \<Longrightarrow> is_perm ip2 [0 ..<length S] \<Longrightarrow>
   map (op ! (map (op ! S) ip1)) ip2 = map (op ! S) (map (op ! ip1) ip2)"
proof(induct ip1 arbitrary: S ip2)
  case Nil
  have S_nil: "S = []" using Nil
       by (metis is_perm_empty_T le_0_eq length_greater_0_conv neq0_conv upt_eq_Nil_conv)
  have ip2_nil: "ip2 = []" using S_nil Nil
       by (metis is_perm_implies_same_length length_greater_0_conv)
  show ?case using Nil S_nil ip2_nil by simp
next
  case (Cons a1 ip1) then show ?case
  apply (simp)
  by (metis Cons.prems(1) in_set_member is_perm_implies_same_length is_perm_member length_map map.simps(2) map_nth member_range nth_map)
qed

lemma upt_Suc_map: "[1 ..<Suc n] = (map Suc [0 ..<n])"
proof(induct n)
  case 0 show ?case by simp
next
  case (Suc n) then show ?case by simp
qed

lemma map_Suc_pred: "(\<forall>x\<in>set S. x > 0) \<Longrightarrow> map Suc (map (\<lambda>x. x - Suc 0) S) = S"
proof(induct S)
  case Nil show ?case by simp
next
  case (Cons a S) then show ?case by simp (* no sledgehammer *)
qed

lemma map_cons_pred: "map (op ! (a#S)) (map Suc ip) = map (op ! S) ip"
proof(induct ip)
  case Nil then show ?case by simp
next
  case (Cons i ip) then show ?case by simp
qed

lemma member_map: "List.member S a \<Longrightarrow> List.member (map f S) (f a)"
proof(induct S)
  case Nil then show ?case by (simp add:List.member_def)
next
  case (Cons b S) then show ?case by (metis map.simps(2) member_rec(1))
qed


(* Removing any occurrence of an element [a] in [S] gives a list that is a permutation
  of removing the first occurrence of [a] in [S]. *)
lemma ip_remove1_split: "i < length S \<Longrightarrow> is_perm (remove1 (S!i) S) (take i S @ drop (Suc i) S)"
proof -
  assume il: "i < length S"
  then have "S = take i S @ S ! i # drop (Suc i) S"
       by (metis id_take_nth_drop)
  then have "is_perm S (S ! i # take i S @ drop (Suc i) S)"
       by (metis is_perm_permute_head_element is_perm_sym)
  then show ?thesis
       by (metis is_perm_cons is_perm_sym)
qed


(* Might be used to ease the proof of [is_perm_comp] below, but it is not the case
  for now (yet this lemma is used later). *)
lemma is_perm_map_nth: "is_perm [0..<length S] ip \<Longrightarrow> i < length S \<Longrightarrow> map (op ! S) ip ! i = S ! (ip ! i)"
by (metis diff_zero is_perm_implies_same_length length_upt nth_map)


(* Take a permutation of the indices of a list, apply it to the list, you get a
  permutation of the original list. *)
lemma is_perm_comp: "is_perm [0..<length S] ip \<Longrightarrow> is_perm S (map (op ! S) ip)"
proof(induct "length S" arbitrary:S ip)
  case 0
  then have ip_nil: "ip = []"
       by (metis is_perm_empty_T upt_0)
  then show ?case using 0 by simp
next
  (* Inductive case. Consider a list of length \<ge> 1, write [S = a # S']. *)
  (* The main argument is that removing any occurrence of an element of a list
    is permutation-equivalent to removing the first occurrence, which is lemma
    [ip_remove1_split]. However, a lot of plumbing is needed here. *)
  case (Suc n)
  obtain a S' where S_decomp: "S = a#S'"
         by (metis Suc.hyps(2) Zero_not_Suc length_0_conv list.exhaust)
  (* Write [rm0] for the list obtained by removing [0] from [ip]. This list is a
    permutation of [Suc 0..<Suc n] and its elements are greater than [0]. *)
  def rm0 \<equiv> "remove1 0 ip"
  have ipS_rm0: "is_perm [Suc 0..<Suc n] rm0"
       by (metis Suc.hyps(2) Suc.prems rm0_def is_perm_upt_Suc_remove0)
  have rm0_seqS: "\<forall>a \<in> set rm0. a \<in> set [Suc 0..<Suc n]"
       by (metis in_set_member ipS_rm0 is_perm_member is_perm_sym)
  have rm0_gt0: "\<forall>a \<in> set [Suc 0..<Suc n]. a > 0"
       by (metis Suc_n_not_le_n neq0_conv in_set_member member_range)
  (* Write [rm0S] for the list obtained by decrementing each element of [rm0] by 1.
    This list is a permutation of [0..<n] and defines a permutation of [S']. *)
  def rm0S \<equiv> "map (\<lambda>x. x - Suc 0) rm0"
  have ipS_ip: "rm0 = map Suc rm0S"
       by (metis map_Suc_pred rm0S_def rm0_seqS rm0_gt0)
  have ip_rm0S: "is_perm [0..<n] rm0S"
       apply (rule_tac f="Suc" in map_is_perm, simp_all)
       by (metis ipS_ip ipS_rm0 map_Suc_upt)
  have IH: "is_perm S' (map (op ! S') rm0S)"
       by (metis S_decomp Suc.hyps ip_rm0S length_Suc_conv list.inject)
  have rm0S_rm0: "map (op ! S') rm0S = map (op ! S) rm0"
       by (metis S_decomp ipS_ip map_cons_pred)
  (* Write [ip0] for the index of the first occurrence of [0] in [ip]. *)
  def ip0 \<equiv> "first_occ 0 ip"
  have a0: "a = S ! 0"
       by (simp add:S_decomp)
  have m0: "List.member ip 0" using Suc
       by (metis is_perm_member member_rec(1) upt_conv_Cons zero_less_Suc)
  have ip0_a: "(map (op ! S) ip) ! ip0 = a"
       by (metis a0 first_occ_member first_occ_member_length ip0_def m0 nth_map)
  (* Removing the first occurrence of [a] or removing any occurrent of [a] give
    equivalent lists. *)
  have ip_S'_rm_aux: "is_perm (remove1 a (map (op ! S) ip)) (take ip0 (map (op ! S) ip) @ drop (Suc ip0) (map (op ! S) ip) )"
       by (metis first_occ_member_length ip0_a ip0_def ip_remove1_split length_map m0)
  have ip_S'_rm: "is_perm (map (op ! S) rm0) (remove1 a (map (op ! S) ip))"
       apply (simp only:rm0_def)
       by (metis drop_map ip0_def ip_S'_rm_aux is_perm_sym m0 map_append remove1_first_occ take_map)

  (* Conclusion *)
  show ?case
       by (metis IH is_perm_cons S_decomp a0 ip_S'_rm is_perm_trans m0 member_map rm0S_rm0)

qed


(* Two following lemmas: not used in the current script, but [is_perm_ex_index_perm]
  gives an interesting relation between permutations of indexes and permutations of
  lists. Helper lemma [member_remove1_split] is actually subsumed by lemma
  [remove1_first_occ] above. It could be removed from the proof of the main lemma
  by defining [i] as [first_occ a T]. *)

lemma member_remove1_split:
  "List.member S a \<Longrightarrow> \<exists>i. 0\<le>i \<and> i<length S \<and> S!i = a \<and> remove1 a S = take i S @ drop (Suc i) S"
by (metis first_occ_member first_occ_member_length le0 remove1_first_occ)

lemma is_perm_ex_index_perm:
  "is_perm S T \<Longrightarrow> \<exists>index_perm. is_perm [0 ..<length T] index_perm \<and>
                                S = map (op ! T) index_perm"
proof(induct S arbitrary: T)
  case Nil
  have "T = []" using Nil.prems by (rule is_perm_empty_T)
  then show ?case by simp
next
  case (Cons a S)
  def T' \<equiv> "remove1 a T"
  (* Obtain an index permutation from the induction hypothesis on [S] *)
  obtain ip
         where ip_cond: "is_perm [0..<length T'] ip \<and> S = map (op ! T') ip"
         using Cons.prems
         by (metis Cons.hyps is_perm_cons T'_def)
  (* Obtain the index at which [a] is removed from [T] *)
  obtain i
         where i_cond: "0\<le>i \<and> i<length T \<and> T!i = a \<and> remove1 a T = take i T @ drop (Suc i) T"
         using Cons.prems
         by (metis is_perm_cons member_remove1_split)
  (* Define an index permutation by combining the two previous elements *)
  def ip_ext \<equiv> "ext_perm i ip"

  (* Now we only need to prove that the index permutation [ip_ext] is
     a valid candidate for the existential quantifier of the case.
     First we state a few helpers... *)
  have min_T_i: "min (length T) i = i"
       by (metis i_cond less_linear less_not_sym min_less_iff_disj)
  have T'_length: "length T = Suc (length T')"
       using Cons.prems T'_def
       by (metis is_perm_cons is_perm_implies_same_length length_Suc_conv)
  have ip_length: "length T = Suc (length ip)"
  proof -
    have "length T' = length [0..<length T']"
         by simp
    also have "... = length ip"
         by(metis is_perm_implies_same_length ip_cond)
    finally show ?thesis using T'_length by simp
  qed
  (* ... then we check the case. *)
  have "is_perm [0..<length T] ip_ext \<and> (a#S) = map (op ! T) ip_ext"
  proof -
    (* First part: [ip_ext] is a permutation of the indexes of [T]. *)
    have "is_perm [0..<length T] ip_ext"
         apply(simp only:ip_ext_def ip_length)
         apply(rule ext_perm_correct[THEN is_perm_sym])
           apply(simp add:i_cond)
          apply (metis i_cond ip_length less_Suc_eq_le)
         by (metis ip_cond ip_length is_perm_sym T'_length nat.inject)
    (* Second part: [ip_ext] relates [T] and [a#S]. *)
    moreover have "(a#S) = map (op ! T) ip_ext"
    proof -
      (* Equality on [a] *)
      have "a = T!i" using i_cond by simp
      (* Equality on the remainder [S] *)
      moreover have "S = map (op ! T \<circ> (lift_from i)) ip"
      proof -
        { (* Quantification on the index *)
          fix x
          assume x_in_ip: "x \<in> set ip"
          (* Another helper *)
          have x_le_T': "x < length T'"
          proof -
            have "List.member [0..<length T'] x"
                 using x_in_ip
                 by (metis ip_cond is_perm_member is_perm_sym List.member_def)
            then show ?thesis
               by (simp add:List.member_def)
          qed
          (* If the index is below the cut *)
          { assume *: "x < i"
            then have "(take i T @ drop (Suc i) T) ! x = T ! x"
                 by (metis append_eq_conv_conj length_take min_T_i nth_take)
          } note case_x_le_i = this
          (* If the index is above the cut *)
          { assume *: "\<not> x < i"
            then have "Suc i + (x - i) \<le> length T" using x_le_T'
                 by (metis Nat.add_diff_inverse T'_length add_Suc_right add_Suc_shift less_imp_le_nat not_less_eq_eq)
            then have "drop (Suc i) T ! (x-i) = T ! ((Suc i) + (x - i))"
                 by(rule nth_drop)
            then have "(take i T @ drop (Suc i) T) ! x = T ! Suc x"
                 by(simp add: nth_append * min_T_i add_diff_inverse)
          } note case_x_not_le_i = this
          (* Putting it together *)
          have "T' ! x = (op ! T \<circ> lift_from i) x"
               using case_x_le_i case_x_not_le_i
               by(simp add:T'_def lift_from_def i_cond)
        } (* Close the local context and generalize [x]. *)

        (* Conclude on the equality on the remainder [S] *)
        then show ?thesis
             by (simp add:ip_cond)
      qed

      (* Equality on the whole list (second part of the goal) *)
      ultimately show ?thesis
                 by(simp add:ip_ext_def ext_perm_def)
    qed

    (* Putting the two parts together *)
    ultimately show ?thesis by blast
  qed

  (* Close the Cons case and the induction *)
  then show ?case
       by(rule exI)
qed



end